/*
 * Copyright (c) 2001-2003 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Gisle Vanem <gvanem@yahoo.no>
 *
 */
#ifndef LWIP_ARCH_CC_H
#define LWIP_ARCH_CC_H

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <conio.h>

#if defined(__DJGPP__) || defined(__MINGW32__)  /* Using MinGW for 'make depend' */
  #include <sys/time.h>
  #define LWIP_TIMEVAL_PRIVATE  0               /* djgpp has its own gettimeofday() */
#else
  struct timeval {
         long  tv_sec;
         long  tv_usec;
       };
  extern int gettimeofday (struct timeval *tv, void *tzone);
#endif

#if 0
/* Define generic types used in lwIP */
typedef unsigned char   u8_t;
typedef signed   char   s8_t;
typedef unsigned short  u16_t;
typedef signed   short  s16_t;
typedef unsigned long   u32_t;
typedef signed   long   s32_t;
#endif

// typedef size_t        mem_ptr_t; // Already defined in $(TOP_DIR)/src/include/lwip/arch.h
typedef unsigned long sys_prot_t;

#if defined(__HIGHC__) || defined(__GNUC__)
  typedef unsigned long long u64_t;

#elif defined(__WATCOMC__) && defined(__WATCOM_INT64__)
  typedef unsigned __int64 u64_t;
#endif

/* Define (sn)printf formatters for these lwIP types */
#define X8_F  "02X"
#define U32_F "lu"
#define S32_F "ld"
#define X32_F "lX"

#if defined(__GNUC__)
  #define U16_F "u"
  #define S16_F "d"
  #define X16_F "X"
  #define SZT_F "u"
#else
  #define U16_F "hu"
  #define S16_F "hd"
  #define X16_F "hX"
  #define SZT_F "lu"
#endif


/*
 * Tricky stuff that depends on compiler and target:
 *  - Macros to peek and poke of DOS-memory (area below 1 MByte). (1)
 *  - Macros to read/write of I/O ports.                          (1)
 *  - Macros to push flags and clear interrrupts.                 (1)
 *  - Inline functions __htonl() and __htons() for swapping
 *    16/32-bit values.                                           (2)
 *  - An alternative fast assembly checksum routine.
 *
 *  (1) only needed when building lwIP.
 *  (2) also used when *using* lwIP.
 */
#if defined(LWIP_BUILDING)

#if defined(__HIGHC__)
  #include <pharlap.h>
  #undef FP_OFF
  #include <dos.h>  /* must come after <pharlap.h> */

  #define PEEKB(s,o)    PeekRealByte (((s) << 16) + (o))
  #define PEEKW(s,o)    PeekRealWord (((s) << 16) + (o))
  #define PEEKL(s,o)    PeekRealDWord(((s) << 16) + (o))
  #define POKEB(s,o,x)  PokeRealByte (((s) << 16) + (o), (x))
  #define POKEW(s,o,x)  PokeRealWord (((s) << 16) + (o), (x))
  #define POKEL(s,o,x)  PokeRealDWord(((s) << 16) + (o), (x))

#elif defined(__DJGPP__)
  #include <dos.h>
  #include <dpmi.h>
  #include <go32.h>
  #include <pc.h>
  #include <sys/farptr.h>

  #define PEEKB(s,o)      _farpeekb (_dos_ds, (o)+((s)<<4))    /**< peek at a u8_t in DOS memory */
  #define PEEKW(s,o)      _farpeekw (_dos_ds, (o)+((s)<<4))    /**< peek at a u16_t in DOS memory */
  #define PEEKL(s,o)      _farpeekl (_dos_ds, (o)+((s)<<4))    /**< peek at a u32_t in DOS memory */
  #define POKEB(s,o,x)    _farpokeb (_dos_ds, (o)+((s)<<4), x) /**< poke a u8_t to DOS memory */
  #define POKEW(s,o,x)    _farpokew (_dos_ds, (o)+((s)<<4), x) /**< poke a u16_t to DOS memory */
  #define POKEL(s,o,x)    _farpokel (_dos_ds, (o)+((s)<<4), x) /**< poke a u32_t to DOS memory */

#elif defined(__WATCOMC__)
  #include <dos.h>
  #include <i86.h>

  #define DOSMEM(s,o,t)   *(volatile t *) (((s) << 4) | (o))
  #define PEEKB(s,o)      DOSMEM(s,o,unsigned char)
  #define PEEKW(s,o)      DOSMEM(s,o,unsigned short)
  #define PEEKL(s,o)      DOSMEM(s,o,unsigned long)
  #define POKEB(s,o,x)    DOSMEM(s,o,unsigned char)  = (unsigned char)(x)
  #define POKEW(s,o,x)    DOSMEM(s,o,unsigned short) = (unsigned short)(x)
  #define POKEL(s,o,x)    DOSMEM(s,o,unsigned long) = (unsigned long)(x)
#endif

#if defined(__HIGHC__)
  #define _inportb(p)     _inb(p)
  #define _inportw(p)     _inw(p)
  #define _outportb(p,x)  _outb(p,x)
  #define _outportw(p,x)  _outpw(p,x)

#elif defined(__DJGPP__)
  #define _inportb(p)     inportb(p)
  #define _inport(p)      inportw(p)
  #define _outportb(p,x)  outportb(p,x)
  #define _outportw(p,x)  outportw(p,x)

#else
  #define _inportb(p)     inp(p)
  #define _inportw(p)     inpw(p)
  #define _outportb(p,x)  outp(p,x)
  #define _outportw(p,x)  outpw(p,x)
#endif

#if defined(__HIGHC__)
  #define PUSHF_CLI()     _inline (0x9C,0xFA)  /* pushfd; cli */
  #define POPF()          _inline (0x9D)       /* popfd */

#elif defined(__GNUC__)
  #define PUSHF_CLI()     __asm__ __volatile__ ("pushfl; cli" ::: "memory")
  #define POPF()          __asm__ __volatile__ ("popfl"       ::: "memory")

#elif defined(__WATCOMC__) && defined(__FLAT__)
  extern void PUSHF_CLI (void);
  #pragma aux PUSHF_CLI = \
          "pushfd"        \
          "cli";

  extern void POPF (void);
  #pragma aux POPF = \
          "popfd";
#endif

#ifndef min
#define min(a,b)  ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a,b)  ((a) > (b) ? (a) : (b))
#endif

#endif  /* LWIP_BUILDING */

/*
 * gcc >= 4.2 have an issue with '__inline__'. We must ensure
 * that always traditional GNU extern inline semantics are used
 * (aka -fgnu89-inline) even if ISO C99 semantics have been specified.
 */
#if defined(__GNUC__) && !defined(__NO_INLINE__)     /* -O0 */
  #define LWIP_GCC_INLINE extern __inline__ __attribute__ ((__gnu_inline__))
#else
  #define LWIP_GCC_INLINE extern __inline__
#endif


#define lwip_htons(x)  __htons(x)
#define lwip_htonl(x)  __htonl(x)

#if defined(__GNUC__) && !defined(__NO_INLINE__)  /* -O0 */

  /* Adapted from <linux/include/asm-386/byteorder.h>.
   */
  LWIP_GCC_INLINE unsigned long __htonl (unsigned long x)
  {
    __asm__ __volatile (
              "xchgb %b0, %h0\n\t"   /* swap lower bytes */
              "rorl  $16, %0\n\t"    /* swap words */
              "xchgb %b0, %h0"       /* swap higher bytes */
            : "=q" (x) : "0" (x));
    return (x);
  }

  LWIP_GCC_INLINE unsigned short __htons (unsigned short x)
  {
    __asm__ __volatile__ (
                "xchgb %b0, %h0"       /* swap bytes */
              : "=q" (x) : "0" (x));
    return (x);
  }

#elif defined(__WATCOMC__) && defined(__FLAT__)  /* wcc386, 32-bit */

  extern unsigned long __htonl (unsigned long x);
  #pragma aux  __htonl =     \
              "xchg al, ah"  \
              "ror  eax, 16" \
              "xchg al, ah"  \
              parm   [eax]   \
              modify [eax];

  extern unsigned short __htons (unsigned short x);
  #pragma aux __htons =     \
              "xchg al, ah" \
              parm   [ax]   \
              modify [ax];

#endif  /* __GNUC__ */

#if !defined(cdecl) && defined(__GNUC__)
#define cdecl __attribute__((__cdecl__))
#endif

#ifndef cdecl
#define cdecl
#endif

/*
 * In chksum0.asm / chksum0.s
 * This asm-version doesn't work with Watcom (don't know why).
 */
extern unsigned short cdecl _in_checksum_fast (const void *ptr, unsigned len);

#if defined(__WATCOMC__)
  /* No decoration. Args on stack. All regs preserved */
  #pragma aux (cdecl) _in_checksum_fast "_*" parm caller[];
#elif defined(__HIGHC__)
  #pragma alias (in_checksum_fast, "_in_checksum_fast")
#endif

#if !defined(__WATCOMC__) && !defined(IN_MAKEFSDATA_C)
  #undef  LWIP_CHKSUM
  #define LWIP_CHKSUM(ptr,len) _in_checksum_fast(ptr,len)
#endif

#define LWIP_PROVIDE_ERRNO
#define BYTE_ORDER  LITTLE_ENDIAN

/* Use arch/bpstruct.h and arch/epstruct.h */
#define PACK_STRUCT_USE_INCLUDES

/* Compiler hints for packing structures */
#if defined(__GNUC__) && 0
  #define PACK_STRUCT_STRUCT     __attribute__((packed))
#else
  #define PACK_STRUCT_STRUCT     /* nop */
#endif

#if defined(__BORLANDC__)
  #include <string.h>           /* Use BCC intrinsic functions */
  #define SMEMCPY(dst,src,len)   __memcpy__(dst,src,len)
#endif

#define LWIP_RAND() ((unsigned long)rand())

#endif /* LWIP_ARCH_CC_H */
