/*
 * Copyright (c) 2001-2003 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: G. Vanem <gvanem@yahoo.no>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <dos.h>
#include <signal.h>

#if defined(__DJGPP__)
#include <sys/time.h>
#endif

#include <lwip/opt.h>
#include <lwip/arch.h>
#include <lwip/stats.h>
#include <lwip/debug.h>
#include <lwip/sys.h>

#include "./pktdrvr.h"
#include "./pkt_if.h"

#if !defined(NO_SYS) || (NO_SYS == 0)
// #error "NO_SYS must be 1 since MSDOS is not able to multitask."
#endif

#ifndef USE_GETTOD
#define USE_GETTOD  1
#endif

#define MAX_JIFFY_OFS ((~0U >> 1)-1)
#define HZ            1000

#define BIOS_CLK 0x46C

#if !defined(__DJGPP__)
int gettimeofday (struct timeval *tv, void *tzone)
{
  struct dostime_t tnow;

  _dos_gettime (&tnow);
  tv->tv_sec  = 3600 * tnow.hour + 60.0 * tnow.minute + tnow.second;
  tv->tv_usec = 10000 * tnow.hsecond;
  return (1);
}
#endif

/*
 * LWIP_DBG_TYPES_ON = 'lwip_debug_flags'
 */
unsigned int lwip_debug_flags;

#if USE_GETTOD
  static struct timeval start_time;
#else
  static u32_t millisec_clock (void);
  static u32_t start_time;
#endif

/*
 * It's vital that we call PktExitDriver() to release the pktdrvr handle.
 * Otherwise the next callback from the driver will crash the DOS-machine
 * (or DOS-box under Windows). The below PktExitDriver() is normally called
 * from pkt_exit(), but calling it twice shouldn't hurt.
 */
static void sig_handler (int sig)
{
  PktExitDriver();

#if defined(__DJGPP__)
  psignal (sig, "Exiting lwIP");
#else
  printf ("Catching signal %d.\n", sig);
#endif

  exit (-1);
}

void sys_init (void)
{
  /* todo: take this from a config-file. E.g. $(LWIP.CFG). + IP, gateway, mask etc.
   */
  const char *env = getenv ("LWIP_DEBUG");

  if (env)
  {
    sscanf (env,"0x%x", &lwip_debug_flags);
    lwip_debug_flags &= ~LWIP_DBG_HALT;
    LWIP_TRACE (0, "debug_flags: 0x%x\n", lwip_debug_flags);
  }

  srand (time(NULL));
#if USE_GETTOD
  gettimeofday (&start_time, NULL);
#else
  start_time = millisec_clock();
#endif

  signal (SIGINT, sig_handler);
  signal (SIGSEGV, sig_handler);
  signal (SIGTERM, sig_handler);
  signal (SIGABRT, sig_handler);
  signal (SIGFPE, sig_handler);
}

/*
 * Return "time now" as expressed by a monotonic increasing milli-sec counter.
 */
u32_t sys_now (void)
{
  return sys_jiffies();
}

/*
 * Return "time now" as expressed by a monotonic increasing milli-sec counter.
 */
u32_t sys_jiffies (void)
{
  struct timeval tv;
  u32_t sec;
  long  usec;

  gettimeofday (&tv, NULL);
  sec = tv.tv_sec - start_time.tv_sec;
  usec = tv.tv_usec;

  if (sec >= (MAX_JIFFY_OFS / HZ))
     return (MAX_JIFFY_OFS);
  usec += (1000000L / HZ) - 1;
  usec /= (1000000L / HZ);
  return (HZ * sec) + usec;
}

/* Provide some real NOP functions too.
 */
#undef sys_arch_protect
#undef sys_arch_unprotect
#undef sys_sem_signal
#undef sys_arch_sem_wait

u32_t sys_arch_protect (void)
{
  return (0);
}

void sys_arch_unprotect (u32_t pval)
{
  LWIP_UNUSED_ARG (pval);
}

void sys_sem_signal (sys_sem_t *sem)
{
  LWIP_UNUSED_ARG (sem);
}

u32_t sys_arch_sem_wait (sys_sem_t *sem, u32_t timeout)
{
  LWIP_UNUSED_ARG (sem);
  LWIP_UNUSED_ARG (timeout);
  return (0xffffffffUL);  // == SYS_ARCH_TIMEOUT
}

#if (USE_GETTOD == 0)
/*
 * The following 2 routines are modified versions from KA9Q NOS
 * by Phil Karn.
 *
 * clockbits() - Read low order bits of timer 0 (the TOD clock)
 * This works only for the 8254 chips used in ATs and 386+.
 *
 * Old version:
 *   The timer 0 runs in mode 3 (square wave mode), counting down
 *   by 2s, twice for each cycle. So it is necessary to read back the
 *   OUTPUT pin to see which half of the cycle we're in. I.e., the OUTPUT
 *   pin forms the most significant bit of the count. Unfortunately,
 *   the 8253 in the PC/XT lacks a command to read the OUTPUT pin...
 *
 * Now simply:
 *   The timer 0 runs in mode 2 (rate generator), counting down.
 *   We don't compare the OUTPUT pin to see what cycle we're in.
 *   A bit lower precision, but a bit faster.
 *
 * Note: The reading of I/O registers gives bad precision under Windows.
 */
static u32_t clockbits (void)
{
  static u32_t count;   /* static because of PUSHF_CLI() */
  static int   started = 0;

  PUSHF_CLI();       /* clear interrupts now */

  if (!started)                  /* Initialize once, it is periodic */
  {
    _outportb (0x43, 0x34);      /* Set timer 0, mode 2 */
    _outportb (0x40, 0x00);      /* Count is decremented, then compared to 0 */
    _outportb (0x40, 0x00);      /* so set count to 0 to get 65536 */
    started = 1;
  }

  _outportb (0x43, 0x04);        /* Latch present counter value command */
  count  = _inportb (0x40);      /* LSB of count */
  count |= _inportb (0x40) << 8; /* MSB of count */

  POPF();
  return (count);
}

/*
 * Return hardware time-of-day in milliseconds. Resolution is improved
 * beyond 55 ms (the clock tick interval) by reading back the instantaneous
 * 8254 counter value and combining it with the clock tick counter.
 *
 * Reading the 8254 is a bit tricky since a tick could occur asynchronously
 * between the two reads in clockbits(). The tick counter is examined before
 * and after the hardware counter is read. If the tick counter changes, try
 * again.
 *
 * Note: The hardware counter (lo) counts down from 65536 at a rate of
 *       1.193181 MHz (4.77/4). System tick count (hi) counts up.
 */
static u32_t millisec_clock (void)
{
  u32_t  hi;
  u16_t  lo;
  double x;

  do
  {
    hi = PEEKL (0, BIOS_CLK);
    lo = clockbits();
  }
  while (hi != PEEKL(0, BIOS_CLK));   /* tick count changed, try again */

  lo = 0 - lo;

  /* Emulating this would be slow. We'd better have a math-processor
   */
  x = ldexp ((double)hi, 16) + (double)lo;  /* x = hi*2^16 + lo */
  return (u32_t) (x * 4.0 / 4770.0);
}
#endif /* USE_GETTOD == 0 */

